//+------------------------------------------------------------------+
//|                                                      MoneyCT.mqh |
//|                   Copyright 2009-2017, MetaQuotes Software Corp. |
//|                                              http://www.mql5.com |
//+------------------------------------------------------------------+
#include <Math\Alglib\dataanalysis.mqh>
#include <Expert\ExpertMoney.mqh>
#include <Trade\DealInfo.mqh>
#include <Math\Stat\Math.mqh>
#include <ct_9.mqh>
// wizard description start
//+------------------------------------------------------------------+
//| Description of the class                                         |
//| Title=Trading with 'Category Theory' Functor optimized volume    |
//| Type=Money                                                       |
//| Name=DiscriminantAnalysis                                        |
//| Class=CMoneyCT                                                   |
//| Page=                                                            |
//| Parameter=DecreaseFactor,double,3.0,Decrease factor              |
//| Parameter=Percent,double,12.5,Percent                            |
//| Parameter=TimeframeOperation,int,0, Timeframe Operation          |
//| Parameter=TimeframeIdentity,int,0, Timeframe Identity            |
//| Parameter=LookbackOperation,int,0, Lookback Operation            |
//| Parameter=LookbackIdentity,int,0, Lookback Identity              |
//| Parameter=AppliedpriceOperation,int,0, Appliedprice Operation    |
//| Parameter=AppliedpriceIdentity,int,0, Appliedprice Identity      |
//| Parameter=IndicatorOperation,int,0, Indicator Operation          |
//| Parameter=IndicatorIdentity,int,0, Indicator Identity            |
//| Parameter=Objects,bool,true,Use Objects to Objects               |
//| Parameter=Decay,double,0.001,Multi Layer Perceptron Decay        |
//| Parameter=Restarts,int,2,Multi Layer Perceptron Restarts         |
//| Parameter=Hidden,int,7,Hidden Layer Size                         |
//+------------------------------------------------------------------+
// wizard description end
//+------------------------------------------------------------------+
//| Class CMoneyCT.                                                  |
//| Purpose: Class of money management with 'Category Theory'        |
//|          Functor optimized volume                                |
//|          optimized trade volume. Derives from class CExpertMoney.|
//+------------------------------------------------------------------+

//+------------------------------------------------------------------+
//| Custom MLP read and write implementations whose soure is not     |
//| shared. Reader can customise their own to their strategy.        |
//+------------------------------------------------------------------+
#import "ct_16.ex5"
  bool Write(string File,double TrainingProfit,CMultilayerPerceptron &MLP);
  bool Read(string File,double TrainingProfit,CMultilayerPerceptron &MLP);
#import

ENUM_TIMEFRAMES                  __TIMEFRAMES[8] = {PERIOD_H1,PERIOD_H2,PERIOD_H3,PERIOD_H4,PERIOD_H6,PERIOD_H8,PERIOD_H12,PERIOD_D1};
int                              __LOOKBACKS[8] = {1,2,3,4,5,6,7,8};
ENUM_APPLIED_PRICE               __APPLIEDPRICES[4] = { PRICE_OPEN, PRICE_MEDIAN, PRICE_TYPICAL, PRICE_CLOSE };
string                           __INDICATORS[2] = { "RSI", "BOLLINGER_BANDS" };
string                           __ACTIONS[2] = { "TRADE_TREND", "TRADE_RANGE" };
#define                          __INPUTS 4
#define                          __OUTPUTS 1
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
class CMoneyCT                   : public CExpertMoney
  {
protected:
   
   double                        m_decrease_factor;
   
   CMonoid<double>               m_timeframe;
   EOperations                   m_timeframe_operation;
   int                           m_timeframe_identity;
   
   CMonoid<double>               m_lookback;
   EOperations                   m_lookback_operation;
   int                           m_lookback_identity;
   
   CMonoid<double>               m_appliedprice;
   EOperations                   m_appliedprice_operation;
   int                           m_appliedprice_identity;
   
   CMonoid<double>               m_indicator;
   EOperations                   m_indicator_operation;
   int                           m_indicator_identity;
   
   bool                          m_objects;        // objects
   double                        m_decay;          // decay
   int                           m_restarts;       // restarts
   int                           m_hidden;         // hidden

public:
                                 CMoneyCT(void);
                                 ~CMoneyCT(void);
                                 
   virtual bool                  ValidationSettings(void);
   //---
   virtual double                CheckOpenLong(double price,double sl);
   virtual double                CheckOpenShort(double price,double sl);
   //---
   void                          DecreaseFactor(double decrease_factor) { m_decrease_factor=decrease_factor;            }
   
   void                          TimeframeOperation(int value)          { m_timeframe_operation=(EOperations)value;     }
   void                          TimeframeIdentity(int value)           { m_timeframe_identity=value;                   }
   
   void                          LookbackOperation(int value)           { m_lookback_operation=(EOperations)value;      }
   void                          LookbackIdentity(int value)            { m_lookback_identity=value;                    }
   
   void                          AppliedpriceOperation(int value)       { m_appliedprice_operation=(EOperations)value;  }
   void                          AppliedpriceIdentity(int value)        { m_appliedprice_identity=value;                }
   
   void                          IndicatorOperation(int value)          { m_indicator_operation=(EOperations)value;     }
   void                          IndicatorIdentity(int value)           { m_indicator_identity=value;                   }
   
   void                          Objects(bool value)                    { m_objects=value;                              }
   void                          Decay(double value)                    { m_decay=value;                                }
   void                          Restarts(int value)                    { m_restarts=value;                             }
   void                          Hidden(int value)                      { m_hidden=value;                               }

protected:

   double                        Optimize(double lots);
   
   CElement<double>              m_element;

   void                          Operate(CMonoid<double> &M,EOperations &O,int IdenityIndex,int &OutputIndex);
   
   ENUM_TIMEFRAMES               SetTimeframe(CMonoid<double> &M,int Index);
   int                           GetTimeframe(ENUM_TIMEFRAMES Timeframe){ int _t=-1; for(int i=0;i<8;i++){ if(__TIMEFRAMES[i]==Timeframe){ _t=i; break; }}  return(_t); }
   
   int                           SetLookback(CMonoid<double> &M,ENUM_TIMEFRAMES Timeframe,int Index);
   int                           GetLookback(int Lookback){ int _l=-1; for(int i=0;i<8;i++){ if(__LOOKBACKS[i]==Lookback){ _l=i; break; }}  return(_l); }
   
   ENUM_APPLIED_PRICE            SetAppliedprice(CMonoid<double> &M,ENUM_TIMEFRAMES Timeframe,int Lookback,int Index);
   int                           GetAppliedprice(ENUM_APPLIED_PRICE Appliedprice){ int _a=-1; for(int i=0;i<4;i++){ if(__APPLIEDPRICES[i]==Appliedprice){ _a=i; break; }}  return(_a); }
   
   double                        SetIndicator(ENUM_TIMEFRAMES Timeframe,int Lookback,ENUM_APPLIED_PRICE Appliedprice,int Index);
   
   int                           RegularizeTimeframe(ENUM_TIMEFRAMES Timeframe){ return(PeriodSeconds(Timeframe)); }
   int                           RegularizeAppliedprice(ENUM_APPLIED_PRICE Appliedprice)
                                 {  if(Appliedprice==PRICE_OPEN){ return(100); }
                                    else if(Appliedprice==PRICE_MEDIAN){ return(200); } 
                                    else if(Appliedprice==PRICE_TYPICAL){ return(300); } 
                                    else if(Appliedprice==PRICE_CLOSE){ return(400); }
                                    return(0); }
   
   CMLPBase                      m_base;
   CMultilayerPerceptron         m_mlp;
   
   bool                          WritePerceptron(double TrainingProfit,CMultilayerPerceptron &MLP);
   bool                          ReadPerceptron(double &TrainingProfit,CMultilayerPerceptron &MLP);
      
   double                        m_training_profit;
   double                        m_training_deposit;
      
  };
//+------------------------------------------------------------------+
//| Constructor                                                      |
//+------------------------------------------------------------------+
void CMoneyCT::CMoneyCT(void)    :  m_decrease_factor(3.0),
                                    m_objects(true),
                                    m_decay(0.001),
                                    m_restarts(2),
                                    m_hidden(7)
  {
//--- initialization of protected data
      m_used_series=USE_SERIES_OPEN+USE_SERIES_HIGH+USE_SERIES_LOW+USE_SERIES_CLOSE+USE_SERIES_SPREAD+USE_SERIES_TIME;
      
      m_timeframe.Cardinality(8);
      m_timeframe.Operation(m_timeframe_operation);
      m_timeframe.Identity(m_timeframe_identity);
      
      m_lookback.Cardinality(8);
      m_lookback.Operation(m_lookback_operation);
      m_lookback.Identity(m_lookback_identity);
      
      m_appliedprice.Cardinality(4);
      m_appliedprice.Operation(m_appliedprice_operation);
      m_appliedprice.Identity(m_appliedprice_identity);
      
      m_indicator.Cardinality(2);
      m_indicator.Operation(m_indicator_operation);
      m_indicator.Identity(m_indicator_identity);
      
      //m_base.MLPCreate1(__INPUTS,m_hidden,__OUTPUTS,m_mlp);
      
//--- 
      m_training_deposit=m_account.Balance();
  }
//+------------------------------------------------------------------+
//| Destructor                                                       |
//+------------------------------------------------------------------+
void CMoneyCT::~CMoneyCT(void)
  {
  }
//+------------------------------------------------------------------+
//| Validation settings protected data.                              |
//+------------------------------------------------------------------+
bool CMoneyCT::ValidationSettings(void)
  {
   if(!CExpertMoney::ValidationSettings())
      return(false);
//--- initial data checks
   if(m_decrease_factor<=0.0)
     {
      printf(__FUNCTION__+": decrease factor must be greater then 0");
      return(false);
     }
     
//--- ok
   return(true);
  }
//+------------------------------------------------------------------+
//| Getting lot size for open long position.                         |
//+------------------------------------------------------------------+
double CMoneyCT::CheckOpenLong(double price,double sl)
  {
      m_time.Refresh(-1);
      
      if(m_time.GetData(0)<=D'2020.02.01')
      {
         if(!ReadPerceptron(m_training_profit,m_mlp))
         {
            m_base.MLPCreate1(__INPUTS,m_hidden,__OUTPUTS,m_mlp);
            m_training_profit=0.0;
         }
         else
         {
            printf(__FUNCSIG__+" read perceptron, with profit: "+DoubleToString(m_training_profit));
         }
      }
      else if(m_time.GetData(0)>=D'2023.07.01')
      {
         printf(__FUNCSIG__+" log profit: "+DoubleToString(m_training_profit)+", account profit: "+DoubleToString(m_account.Profit())+", equity: "+DoubleToString(m_account.Equity())+", deposit: "+DoubleToString(m_training_deposit));
         
         if(m_training_profit<m_account.Equity()-m_training_deposit)
         {
            printf(__FUNCSIG__+" perceptron write... ");
            m_training_profit=m_account.Equity()-m_training_deposit;
            
            WritePerceptron(m_training_profit,m_mlp);
         }
      }
      
      if(m_symbol==NULL){
         return(0.0); }
      
      ENUM_TIMEFRAMES _timeframe_1=SetTimeframe(m_timeframe,1);
      int _lookback_1=SetLookback(m_lookback,_timeframe_1,1);
      ENUM_APPLIED_PRICE _appliedprice_1=SetAppliedprice(m_appliedprice,_timeframe_1,_lookback_1,1);
      double _indicator_1=SetIndicator(_timeframe_1,_lookback_1,_appliedprice_1,1);
      
      m_open.Refresh(-1);
      m_low.Refresh(-1);
      m_close.Refresh(-1);
      
      CMLPTrain _train;
      
      int _info=0;
      CMLPReport _report;
      CMatrixDouble _xy;_xy.Resize(1,__INPUTS+__OUTPUTS);
      
      _xy[0].Set(0,RegularizeTimeframe(_timeframe_1));
      _xy[0].Set(1,_lookback_1);
      _xy[0].Set(2,RegularizeAppliedprice(_appliedprice_1));
      _xy[0].Set(3,_indicator_1);
      //
      int _x=StartIndex()+1;
      
      double _sl_1=m_high.GetData(_x)-m_low.GetData(_x);
      
      if(m_open.GetData(_x)<m_close.GetData(_x))
      {
         _sl_1=m_open.GetData(_x)-m_low.GetData(_x);
      }
      
      double _stops=(2.0*(m_symbol.Ask()-m_symbol.Bid()))+((m_symbol.StopsLevel()+m_symbol.FreezeLevel())*m_symbol.Point());
      
      _xy[0].Set(__INPUTS,fmax(_stops,_sl_1));
      
      _train.MLPTrainLM(m_mlp,_xy,1,m_decay,m_restarts,_info,_report);
      
      //printf(__FUNCSIG__+" sl 1: "+DoubleToString(_sl_1)+" stops: "+DoubleToString(_stops));
      
      if(_info>0)
      {
         ENUM_TIMEFRAMES _timeframe_0=SetTimeframe(m_timeframe,0);
         int _lookback_0=SetLookback(m_lookback,_timeframe_0,0);
         ENUM_APPLIED_PRICE _appliedprice_0=SetAppliedprice(m_appliedprice,_timeframe_0,_lookback_0,0);
         double _indicator_0=SetIndicator(_timeframe_0,_lookback_0,_appliedprice_0,0);
         
         double _x_inputs[],_y_output[];
         ArrayResize(_x_inputs,__INPUTS);
         ArrayResize(_y_output,__OUTPUTS);
         
         ArrayInitialize(_x_inputs,0.0);
         ArrayInitialize(_y_output,0.0);
         
         _x_inputs[0]=RegularizeTimeframe(_timeframe_0);
         _x_inputs[1]=_lookback_0;
         _x_inputs[2]=RegularizeAppliedprice(_appliedprice_0);
         _x_inputs[3]=_indicator_0;
         
         //forward feed
         m_base.MLPProcess(m_mlp,_x_inputs,_y_output);
         
         double _stoploss_gap=_y_output[0];
         
         //printf(__FUNCSIG__+" ct call: "+DoubleToString(_stoploss_gap));
         
         sl=m_symbol.Bid()-fabs(_stoploss_gap);
         
         //--- select lot size
         double _ct_1_lot_loss=(_stoploss_gap/m_symbol.TickSize())*m_symbol.TickValue();
         double lot=((m_percent/100.0)*m_account.FreeMargin())/_ct_1_lot_loss;
         
         //--- calculate margin requirements for 1 lot
         if(m_account.FreeMarginCheck(m_symbol.Name(),ORDER_TYPE_BUY,lot,m_symbol.Ask())<0.0)
         {
            printf(__FUNCSIG__" insufficient margin for sl lot! ");
            lot=m_account.MaxLotCheck(m_symbol.Name(),ORDER_TYPE_BUY,m_symbol.Ask(),m_percent);
         }
         
         //--- return trading volume
         return(Optimize(lot));
      }
      
      return(0.0);
  }
//+------------------------------------------------------------------+
//| Getting lot size for open short position.                        |
//+------------------------------------------------------------------+
double CMoneyCT::CheckOpenShort(double price,double sl)
  {
      if(m_time.GetData(0)<=D'2020.02.01')
      {
         if(!ReadPerceptron(m_training_profit,m_mlp))
         {
            m_base.MLPCreate1(__INPUTS,m_hidden,__OUTPUTS,m_mlp);
            m_training_profit=0.0;
         }
         else
         {
            printf(__FUNCSIG__+" read perceptron, with profit: "+DoubleToString(m_training_profit));
         }
      }
      else if(m_time.GetData(0)>=D'2023.07.01')
      {
         printf(__FUNCSIG__+" log profit: "+DoubleToString(m_training_profit)+", account profit: "+DoubleToString(m_account.Profit())+", equity: "+DoubleToString(m_account.Equity())+", deposit: "+DoubleToString(m_training_deposit));
         
         if(m_training_profit<m_account.Equity()-m_training_deposit)
         {
            printf(__FUNCSIG__+" perceptron write... ");
            m_training_profit=m_account.Equity()-m_training_deposit;
            
            WritePerceptron(m_training_profit,m_mlp);
         }
      }
      
      if(m_symbol==NULL){
         return(0.0); }
      
      ENUM_TIMEFRAMES _timeframe_1=SetTimeframe(m_timeframe,1);
      int _lookback_1=SetLookback(m_lookback,_timeframe_1,1);
      ENUM_APPLIED_PRICE _appliedprice_1=SetAppliedprice(m_appliedprice,_timeframe_1,_lookback_1,1);
      double _indicator_1=SetIndicator(_timeframe_1,_lookback_1,_appliedprice_1,1);
      
      m_open.Refresh(-1);
      m_high.Refresh(-1);
      m_close.Refresh(-1);
      
      CMLPTrain _train;
      
      int _info=0;
      CMLPReport _report;
      CMatrixDouble _xy;_xy.Resize(1,__INPUTS+__OUTPUTS);
      
      _xy[0].Set(0,RegularizeTimeframe(_timeframe_1));
      _xy[0].Set(1,_lookback_1);
      _xy[0].Set(2,RegularizeAppliedprice(_appliedprice_1));
      _xy[0].Set(3,_indicator_1);
      //
      int _x=StartIndex()+1;
      
      double _sl_1=m_high.GetData(_x)-m_low.GetData(_x);
      
      if(m_open.GetData(_x)>m_close.GetData(_x))
      {
         _sl_1=m_high.GetData(_x)-m_open.GetData(_x);
      }
      
      double _stops=(2.0*(m_symbol.Ask()-m_symbol.Bid()))+((m_symbol.StopsLevel()+m_symbol.FreezeLevel())*m_symbol.Point());
      
      _xy[0].Set(__INPUTS,fmax(_stops,_sl_1));
      
      _train.MLPTrainLM(m_mlp,_xy,1,m_decay,m_restarts,_info,_report);
      
      //printf(__FUNCSIG__+" sl 1: "+DoubleToString(_sl_1)+" stops: "+DoubleToString(_stops));
      
      if(_info>0)
      {
         ENUM_TIMEFRAMES _timeframe_0=SetTimeframe(m_timeframe,0);
         int _lookback_0=SetLookback(m_lookback,_timeframe_0,0);
         ENUM_APPLIED_PRICE _appliedprice_0=SetAppliedprice(m_appliedprice,_timeframe_0,_lookback_0,0);
         double _indicator_0=SetIndicator(_timeframe_0,_lookback_0,_appliedprice_0,0);
         
         double _x_inputs[],_y_output[];
         ArrayResize(_x_inputs,__INPUTS);
         ArrayResize(_y_output,__OUTPUTS);
         
         ArrayInitialize(_x_inputs,0.0);
         ArrayInitialize(_y_output,0.0);
         
         _x_inputs[0]=RegularizeTimeframe(_timeframe_0);
         _x_inputs[1]=_lookback_0;
         _x_inputs[2]=RegularizeAppliedprice(_appliedprice_0);
         _x_inputs[3]=_indicator_0;
         
         //forward feed?...
         m_base.MLPProcess(m_mlp,_x_inputs,_y_output);
         
         //output from MLP forecast
         double _stoploss_gap=_y_output[0];
         
         //printf(__FUNCSIG__+" ct call: "+DoubleToString(_stoploss_gap));
      
         sl=m_symbol.Ask()+fabs(_stoploss_gap);
         
         //--- select lot size
         double _ct_1_lot_loss=(_stoploss_gap/m_symbol.TickSize())*m_symbol.TickValue();
         double lot=((m_percent/100.0)*m_account.FreeMargin())/_ct_1_lot_loss;
         
         //--- calculate margin requirements for 1 lot
         if(m_account.FreeMarginCheck(m_symbol.Name(),ORDER_TYPE_SELL,lot,m_symbol.Bid())<0.0)
         {
            printf(__FUNCSIG__" insufficient margin for sl lot! ");
            lot=m_account.MaxLotCheck(m_symbol.Name(),ORDER_TYPE_SELL,m_symbol.Bid(),m_percent);
         }
         
         //--- return trading volume
         return(Optimize(lot));
      }
      
      return(0.0);
  }
//+------------------------------------------------------------------+
//| Optimizing lot size for open.                                    |
//+------------------------------------------------------------------+
double CMoneyCT::Optimize(double lots)
  {
   double lot=lots;
//--- calculate number of losses orders without a break
   if(m_decrease_factor>0)
     {
      //--- select history for access
      HistorySelect(0,TimeCurrent());
      //---
      int       orders=HistoryDealsTotal();  // total history deals
      int       losses=0;                    // number of consequent losing orders
      CDealInfo deal;
      //---
      for(int i=orders-1;i>=0;i--)
        {
         deal.Ticket(HistoryDealGetTicket(i));
         if(deal.Ticket()==0)
           {
            Print("CMoneyCT::Optimize: HistoryDealGetTicket failed, no trade history");
            break;
           }
         //--- check symbol
         if(deal.Symbol()!=m_symbol.Name())
            continue;
         //--- check profit
         double profit=deal.Profit();
         if(profit>0.0)
            break;
         if(profit<0.0)
            losses++;
        }
      //---
      if(losses>1)
         lot=NormalizeDouble(lot-lot*losses/m_decrease_factor,2);
     }
//--- normalize and check limits
   double stepvol=m_symbol.LotsStep();
   lot=stepvol*NormalizeDouble(lot/stepvol,0);
//---
   double minvol=m_symbol.LotsMin();
   if(lot<minvol)
      lot=minvol;
//---
   double maxvol=m_symbol.LotsMax();
   if(lot>maxvol)
      lot=maxvol;
//---
   return(lot);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void CMoneyCT::Operate(CMonoid<double> &M,EOperations &O,int IdenityIndex,int &OutputIndex)
   {
      OutputIndex=-1;
      //
      double _values[];
      ArrayResize(_values,M.Cardinality());ArrayInitialize(_values,0.0);
      //
      for(int i=0;i<M.Cardinality();i++)
      {
         m_element.Let();
         if(M.Get(i,m_element))
         {
            if(!m_element.Get(0,_values[i]))
            {
               printf(__FUNCSIG__+" Failed to get double for 1 at: "+IntegerToString(i+1));
            }
         }
         else{ printf(__FUNCSIG__+" Failed to get element for 1 at: "+IntegerToString(i+1)); }
      }
      
      //
      
      if(O==OP_LEAST)
      {
         OutputIndex=0;
         double _least=_values[0];
         for(int i=0;i<M.Cardinality();i++)
         {
            if(_least>_values[i] || i==IdenityIndex){ _least=_values[i]; OutputIndex=i; }
         }
      }
      else if(O==OP_MOST)
      {
         OutputIndex=0;
         double _most=_values[0];
         for(int i=0;i<M.Cardinality();i++)
         {
            if(_most<_values[i] || i==IdenityIndex){ _most=_values[i]; OutputIndex=i; }
         }
      }
      else if(O==OP_CLOSEST)
      {
         double _mean=0.0;
         for(int i=0;i<M.Cardinality();i++)
         {
            _mean+=_values[i];
         }
         _mean/=M.Cardinality();
         
         OutputIndex=0;
         double _closest=fabs(_values[0]-_mean);
         for(int i=0;i<M.Cardinality();i++)
         {
            if(_closest>fabs(_values[i]-_mean) || i==IdenityIndex){ _closest=fabs(_values[i]-_mean); OutputIndex=i; }
         }
      }
      else if(O==OP_FURTHEST)
      {
         double _mean=0.0;
         for(int i=0;i<M.Cardinality();i++)
         {
            _mean+=_values[i];
         }
         _mean/=M.Cardinality();
         
         OutputIndex=0;
         double _furthest=fabs(_values[0]-_mean);
         for(int i=0;i<M.Cardinality();i++)
         {
            if(_furthest<fabs(_values[i]-_mean) || i==IdenityIndex){ _furthest=fabs(_values[i]-_mean); OutputIndex=i; }
         }
      }
   }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
ENUM_TIMEFRAMES CMoneyCT::SetTimeframe(CMonoid<double> &M,int Index)
   {
      for(int i=0;i<8;i++)
      {
         ResetLastError();
         double _value=0.0;
         double _buffer[];ArrayResize(_buffer,3);ArrayInitialize(_buffer,0.0);ArraySetAsSeries(_buffer,true);
         if(CopyClose(m_symbol.Name(),__TIMEFRAMES[i],Index,3,_buffer)>=3)
         {
            _value=fabs(_buffer[0]-_buffer[1])/(fabs(_buffer[0]-_buffer[1])+fabs(_buffer[1]-_buffer[2]));
         }
         else{ printf(__FUNCSIG__+" Failed to copy: "+EnumToString(__TIMEFRAMES[i])+", at index: "+IntegerToString(i)+" close prices. err: "+IntegerToString(GetLastError())); }
         
         m_element.Let();
         m_element.Cardinality(1);
         if(m_element.Set(0,_value))
         {
            ResetLastError();
            if(!M.Set(i,m_element,false))
            {
               printf(__FUNCSIG__+" Failed to assign element at index: "+IntegerToString(i)+", for lookback. ERR: "+IntegerToString(GetLastError()));
            }
         }
      }
      
      int _i_out=-1;
      //
      Operate(M,m_timeframe_operation,m_timeframe_identity,_i_out);
      
      if(_i_out==-1){ return(INVALID_HANDLE); }
      
      return(__TIMEFRAMES[_i_out]);
   }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
int CMoneyCT::SetLookback(CMonoid<double> &M,ENUM_TIMEFRAMES Timeframe,int Index)
   {
      m_close.Refresh(-1);
      
      int _x=StartIndex()+Index;
      
      for(int i=0;i<8;i++)
      {
         int _period=(__LOOKBACKS[i]*PeriodSeconds(Timeframe))/PeriodSeconds(m_period);
         double _value=fabs(m_close.GetData(_x)-m_close.GetData(_x+_period))/(fabs(m_close.GetData(_x)-m_close.GetData(_x+_period))+fabs(m_close.GetData(_x+_period)-m_close.GetData(_x+_period+_period)));
         
         m_element.Let();
         m_element.Cardinality(1);
         if(m_element.Set(0,_value))
         {
            ResetLastError();
            if(!M.Set(i,m_element,false))
            {
               printf(__FUNCSIG__+" Failed to assign element at index: "+IntegerToString(i)+", for lookback. ERR: "+IntegerToString(GetLastError()));
            }
         }
      }
   
      int _i_out=-1;
      //
      Operate(M,m_lookback_operation,m_lookback_identity,_i_out);
      
      if(_i_out==-1){ return(4); }
      
      return(4*__LOOKBACKS[_i_out]);
   }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
ENUM_APPLIED_PRICE CMoneyCT::SetAppliedprice(CMonoid<double> &M,ENUM_TIMEFRAMES Timeframe,int Lookback,int Index)
   {
      for(int i=0;i<4;i++)
      {
         ResetLastError();
         double _value=0.0;
         if(__APPLIEDPRICES[i]==PRICE_CLOSE)
         {
            double _buffer[];ArrayResize(_buffer,Lookback);ArrayInitialize(_buffer,0.0);ArraySetAsSeries(_buffer,true);
            if(CopyClose(m_symbol.Name(),Timeframe,Index,Lookback,_buffer)>=Lookback)
            {
               _value=MathStandardDeviation(_buffer);
            }
            else{ printf(__FUNCSIG__+" Failed to copy: "+EnumToString(Timeframe)+", at index: "+IntegerToString(i)+" close prices. err: "+IntegerToString(GetLastError())); }
         }
         else if(__APPLIEDPRICES[i]==PRICE_OPEN)
         {
            double _buffer[];ArrayResize(_buffer,Lookback);ArrayInitialize(_buffer,0.0);ArraySetAsSeries(_buffer,true);
            if(CopyOpen(m_symbol.Name(),Timeframe,Index,Lookback,_buffer)>=Lookback)
            {
               _value=MathStandardDeviation(_buffer);
            }
            else{ printf(__FUNCSIG__+" Failed to copy: "+EnumToString(Timeframe)+", at index: "+IntegerToString(i)+" open prices. err: "+IntegerToString(GetLastError())); }
         }
         else if(__APPLIEDPRICES[i]==PRICE_MEDIAN)
         {
            double _buffer[];ArrayResize(_buffer,Lookback);ArrayInitialize(_buffer,0.0);ArraySetAsSeries(_buffer,true);
            double _h[];ArrayResize(_h,Lookback);ArrayInitialize(_h,0.0);
            double _l[];ArrayResize(_l,Lookback);ArrayInitialize(_l,0.0);
            if(CopyHigh(m_symbol.Name(),Timeframe,Index,Lookback,_h)>=Lookback && CopyLow(m_symbol.Name(),Timeframe,Index,Lookback,_l)>=Lookback)
            {
               for(int ii=0;ii<Lookback;ii++)
               {
                  _buffer[ii]=0.5*(_h[ii]+_l[ii]);
               }
               _value=MathStandardDeviation(_buffer);
            }
            else{ printf(__FUNCSIG__+" Failed to copy: "+EnumToString(Timeframe)+", at index: "+IntegerToString(i)+" high and/or low prices. err: "+IntegerToString(GetLastError())); }
         }
         else if(__APPLIEDPRICES[i]==PRICE_TYPICAL)
         {
            double _buffer[];ArrayResize(_buffer,Lookback);ArrayInitialize(_buffer,0.0);ArraySetAsSeries(_buffer,true);
            double _h[];ArrayResize(_h,Lookback);ArrayInitialize(_h,0.0);
            double _l[];ArrayResize(_l,Lookback);ArrayInitialize(_l,0.0);
            double _c[];ArrayResize(_c,Lookback);ArrayInitialize(_c,0.0);
            if(CopyHigh(m_symbol.Name(),Timeframe,Index,Lookback,_h)>=Lookback && CopyLow(m_symbol.Name(),Timeframe,Index,Lookback,_l)>=Lookback && CopyClose(m_symbol.Name(),Timeframe,0,Lookback,_c)>=Lookback)
            {
               for(int ii=0;ii<Lookback;ii++)
               {
                  _buffer[ii]=(1.0/3.0)*(_h[ii]+_l[ii]+_c[ii]);
               }
               _value=MathStandardDeviation(_buffer);
            }
            else{ printf(__FUNCSIG__+" Failed to copy: "+EnumToString(Timeframe)+", at index: "+IntegerToString(i)+" high and/or low prices. err: "+IntegerToString(GetLastError())); }
         }
         else if(__APPLIEDPRICES[i]==PRICE_WEIGHTED)
         {
            double _buffer[];ArrayResize(_buffer,Lookback);ArrayInitialize(_buffer,0.0);ArraySetAsSeries(_buffer,true);
            double _h[];ArrayResize(_h,Lookback);ArrayInitialize(_h,0.0);
            double _l[];ArrayResize(_l,Lookback);ArrayInitialize(_l,0.0);
            double _c[];ArrayResize(_c,Lookback);ArrayInitialize(_c,0.0);
            if(CopyHigh(m_symbol.Name(),Timeframe,Index,Lookback,_h)>=Lookback && CopyLow(m_symbol.Name(),Timeframe,Index,Lookback,_l)>=Lookback && CopyClose(m_symbol.Name(),Timeframe,Index,Lookback,_c)>=Lookback)
            {
               for(int ii=0;ii<Lookback;ii++)
               {
                  _buffer[ii]=(1.0/4.0)*(_h[ii]+_l[ii]+_c[ii]+_c[ii]);
               }
               _value=MathStandardDeviation(_buffer);
            }
            else{ printf(__FUNCSIG__+" Failed to copy: "+EnumToString(Timeframe)+", at index: "+IntegerToString(i)+" high and/or low prices. err: "+IntegerToString(GetLastError())); }
         }
         
         m_element.Let();
         m_element.Cardinality(1);
         if(m_element.Set(0,_value))
         {
            ResetLastError();
            if(!M.Set(i,m_element,false))
            {
               printf(__FUNCSIG__+" Failed to assign element at index: "+IntegerToString(i)+", for applied price. ERR: "+IntegerToString(GetLastError()));
            }
         }
      }
      
      int _i_out=-1;
      //
      Operate(M,m_appliedprice_operation,m_appliedprice_identity,_i_out);
      
      if(_i_out==-1){ return(4); }
      
      return(__APPLIEDPRICES[_i_out]);
   }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
double CMoneyCT::SetIndicator(ENUM_TIMEFRAMES Timeframe,int Lookback,ENUM_APPLIED_PRICE Appliedprice,int Index)
   {
      m_spread.Refresh(-1);
      
      int _x=StartIndex();
      
      double _indicator[2];
      ArrayInitialize(_indicator,0.0);
      
      for(int i=0;i<2;i++)
      {
         ResetLastError();
         double _value=0.0;
         //
         if(__INDICATORS[i]=="RSI")
         {
            double _spread=(m_symbol.Point()*m_spread.GetData(_x));
            double _current_gain=_spread,_current_loss=_spread;
            double _gain=_spread,_loss=_spread;
            double _buffer[];ArrayResize(_buffer,Lookback);ArrayInitialize(_buffer,0.0);ArraySetAsSeries(_buffer,true);
            if(CopyClose(m_symbol.Name(),Timeframe,Index,Lookback,_buffer)>=Lookback)
            {
               if(_buffer[0]>_buffer[1]){ _current_gain+=_buffer[0]-_buffer[1]; }
               else if(_buffer[1]>_buffer[0]){ _current_loss+=_buffer[1]-_buffer[0]; }
               
               for(int ii=1;ii<Lookback-1;ii++)
               {
                  if(_buffer[ii]>_buffer[ii+1]){ _gain+=_buffer[ii]-_buffer[ii+1]; }
                  else if(_buffer[ii+1]>_buffer[ii]){ _loss+=_buffer[ii+1]-_buffer[ii]; }
               }
               
               _gain/=(Lookback-1);
               _loss/=(Lookback-1);
               
               _gain+=_current_gain;
               _loss+=_current_loss;
               
               _value=(_gain/_loss);
               _value+=1.0;
               _value=100.0/_value;
               _value=100.0-_value;
               
               _indicator[0]=_value;
            }
            else{ printf(__FUNCSIG__+" Failed to copy: "+EnumToString(Timeframe)+" close prices. err: "+IntegerToString(GetLastError())); }
         }
         else if(__INDICATORS[i]=="BOLLINGER_BANDS")
         {
            double _base=0.0,_upper=0.0,_lower=0.0;
            double _buffer[];ArrayResize(_buffer,Lookback);ArrayInitialize(_buffer,0.0);ArraySetAsSeries(_buffer,true);
            if(CopyClose(m_symbol.Name(),Timeframe,Index,Lookback,_buffer)>=Lookback)
            {
               for(int ii=0;ii<Lookback;ii++)
               {
                  _base+=_buffer[ii];
               }
               
               _base/=(Lookback);
               
               double _deviation=MathStandardDeviation(_buffer);
               
               _upper+=(_base+(2.0*_deviation));
               _lower+=(_base-(2.0*_deviation));
               
               _value=((1.0+((_buffer[0]-_base)/(fabs(_buffer[0]-_base)+(_upper-_lower))))*50.0);
               
               _indicator[1]=_value;
            }
            else{ printf(__FUNCSIG__+" Failed to copy: "+EnumToString(Timeframe)+" close prices. err: "+IntegerToString(GetLastError())); }
         }
         
         //
         
         m_element.Let();
         m_element.Cardinality(1);
         if(m_element.Set(0,_value))
         {
            ResetLastError();
            if(!m_indicator.Set(i,m_element,false))
            {
               printf(__FUNCSIG__+" Failed to assign element at index: "+IntegerToString(i)+", for indicator. ERR: "+IntegerToString(GetLastError()));
            }
         }
      }
      
      int _i_out=-1;
      Operate(m_indicator,m_indicator_operation,m_indicator_identity,_i_out);
      
      return(_indicator[_i_out]);
   }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool CMoneyCT::WritePerceptron(double TrainingProfit,CMultilayerPerceptron &MLP)
   {
      string _file="_s_perceptron_"+m_account.Currency()+"_"+m_symbol.Name()+"_"+EnumToString(m_period)+"_"+IntegerToString(m_hidden)+"_"+string(m_objects)+".bin";
      
      return(Write(_file,TrainingProfit,MLP));
   }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool CMoneyCT::ReadPerceptron(double &TrainingProfit,CMultilayerPerceptron &MLP)
   {
      string _file="_s_perceptron_"+m_account.Currency()+"_"+m_symbol.Name()+"_"+EnumToString(m_period)+"_"+IntegerToString(m_hidden)+"_"+string(m_objects)+".bin";
      
      return(Read(_file,TrainingProfit,MLP));
   }
//+------------------------------------------------------------------+
